# $Id: azat.awk,v 1.12 2008/03/07 08:57:01 heinz Exp $
#
# AZAT - AWK Zeit Aufschreibungs Tool
# Copyright (C) 2004-2006  Heinz Deinhart
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; only version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-13

### data structures:
# active_day  -> stores the active day while parsing
# init_month_db[YYMM] = 0|1  -> stores what months are initialized
# month_db[<month_counter>]  -> stores what months are stored
# should_db[YYYYMMDD]  -> stores the planned working time in minutes
# time_db[YYYYMMDD]  -> stores the actually worked time in minutes
# holiday_db[YYYYMMDD]  -> stores holiday amount of day in percent
# holidays_begin[YYYYMM] -> stores holidays at begin of month
# holidays_end[YYYYMM] -> stores holidays at begin of month
# month_time[YYYYMM]  -> stores the sum of times
# month_should[YYYYMM]  -> stores the sum of shoulds
# month_diff[YYYYMM]  -> stores the diff for this month
# month_diffsum[YYYYMM]  -> stores the sum of all diffs
# job_db[YYYYMMDD,<job_db_counter[YYYMMDD]>]  -> stores the job description
# job_db_counter[YYYYMMDD] -> stores job desc counter
# errors[error_counter] -> stores error or warning messages
# comment_db[YYYYMMDD] -> stores comments for days

#TODO
# falsche zeiten anschaun
# dulpizierte ausgaben checken

function azat_error( str ) {
  errors[error_counter++] = "Error: " str;
}
function azat_warning( str ) {
  errors[error_counter++] = "Warning: " str;
}

function fatal_error( str ) {
    printf "fatal error: %s\n", str > /dev/stderr;
    exit 1
}

function date_launch( str, fmt, lang,
		      cmd, result ) {
    cmd = sprintf( "%s date -d %s +'%s' 2>/dev/null || echo illegal-date", lang, str, fmt );
    cmd | getline result;
    close( cmd );
    return result;
}

function date_get_ym( str ) {
    return date_launch( str, "%Y%m" );
}

function date_get_day_name_long( str ) {
    return date_launch( str, "%A", date_launch_env );
}

function date_get_day_name_short( str ) {
    return date_launch( str, "%a", date_launch_env );
}

function date_get_day_of_week( str ) {
    return date_launch( str, "%u" );
}

function date_get_pretty_day( str ) {
    return date_launch(  str, "%e. %b, %Y", date_launch_env );
}

function date_get_pretty_month( str ) {
    return date_launch(  sprintf( "%s01", str ), "%B %Y", date_launch_env );
}

function date_get_year( str ) {
    return date_launch( sprintf( "%s01", str ), "%Y" );
}

function date_get_month_name_long( str ) {
    return date_launch( sprintf( "%s01", str ), "%B", date_launch_env );
}

function date_get_monthlength( ym ) {
    if (dbg) printf "date_get_monthlength(%s)\n", ym;
    if (date_launch( ym"29", "%d" ) == "illegal-date")
	return 28;
    if (date_launch( ym"30", "%d" ) == "illegal-date")
	return 29;
    if (date_launch( ym"31", "%d" ) == "illegal-date")
	return 30;
    if (date_launch( ym"32", "%d" ) == "illegal-date")
	return 31;

    printf "fatal error: date_get_monthlength fucked up\n"
    return 0;
}

function is_weekend( str,
		     indexx ) {
    indexx = date_launch( str, "%w" );
    if (indexx < 1 || indexx > 5)
	return 1;
    return 0;
}

function init_month( date,
		     ym, ml, day, the_day ) {
    ym = date_get_ym( date );
    if (init_month_db[ym] == 1)
	return 0;

    if (dbg) printf "init_month(%s): ym=%s; month_counter=%i\n", 
	date, ym, month_counter;
    month_db[month_counter++] = ym;

    ml = date_get_monthlength( ym );
    if(dbg) printf "now initialising month %s of length %i\n", date, ml;
    for (day = 1; day <= ml; day++) {
	if (dbg) printf "initialising day %i of %s\n", day, date;
	the_day = sprintf( "%i%02i", ym, day );
	if (is_weekend( the_day )) # weekend
	    should_db[the_day] = 0;
	else # workday
	    should_db[the_day] = 8 * 60;
    }

    init_month_db[ym] = 1;
}

function str2time( str,
		   t ) {
    split( str, t, ":" );
    return t[1] * 60 + t[2];
}

function min2str( mins,
                  hours, negc ) {
    if (mins < 0) {
        negc = "-";
        mins = -1 * mins;
    } else
        negc = " ";
    hours = 0;
    while (mins >= 60) {
        hours++;
        mins -= 60;
    }
    return sprintf( "%s%2i:%02i", negc, hours, mins );
}

function calc_month_holidays( ym, last_ym,
			      ml, day, the_day) {
    if (dbg) printf "called calc_month_holidays( %s )\n", ym;
    holidays_begin[ym] += holidays_end[last_ym];
    holidays_end[ym] = holidays_begin[ym] * 100;
    ml = date_get_monthlength( ym );
    for (day = 1; day <= ml; day++) {
	the_day = sprintf( "%i%02i", ym, day );
	if (holiday_db[the_day] > 0) {
	    holidays_end[ym] -= holiday_db[the_day];
	    should_db[the_day] *= (100 - holiday_db[the_day]);
	    should_db[the_day] /= 100;
	}
    }
    holidays_end[ym] /= 100;
}

function calc_month_timetable( ym, last_ym, diffsumadder,
			       the_day, ml ) {
    if (dbg) printf "called calc_month_timetable( %s )\n", ym;
    ml = date_get_monthlength( ym );
    month_should[ym] = 0;
    month_time[ym] = 0;
    month_diff[ym] = 0;

    for (day = 1; day <= ml; day++) {
	the_day = sprintf( "%i%02i", ym, day );
	month_should[ym] += should_db[the_day];
	month_time[ym] += time_db[the_day];
	month_diff[ym] += time_db[the_day] - should_db[the_day];
    }

    month_diffsum[ym] = month_diff[ym] + month_diffsum[last_ym] + diffsumadder;
}

/./ {
    consumed_line = 0;
}

/^[0-9]+[[:blank:]]+[0-9]+:[0-9]+[[:blank:]]+[0-9]+:[0-9]+[[:blank:]]*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    active_day = _the_day;
    _i1 = str2time( $2 );
    _i2 = str2time( $3 );
    if (_i2 <= _i1)
	fatal_error( "illegal timestamps" );
    if (dbg) printf "stores for %s\n", _the_day;
    time_db[_the_day] = time_db[_the_day] + _i2 - _i1;
}

/^[0-9]+[[:blank:]]+(FEIERTAG|PUBLICHOLIDAY)[[:blank:]]*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    should_db[_the_day] = 0;
    comment_db[_the_day] = tl_pubholiday;
}

/^[0-9]+[[:blank:]]+(SPEZIALSOLL|CUSTOMSHOULD)[[:blank:]]+[01-9]+[[:blank:]]*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    should_db[_the_day] = $3 * 60;
    comment_db[_the_day] = sprintf( "%s=%ih", tl_should, $3 );
}

/^[0-9]+[[:blank:]]+(SPEZIALSOLL|CUSTOMSHOULD)[[:blank:]]+[01-9]+[[:blank:]]+[Mm][Ii][Nn][[:blank:]]*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    should_db[_the_day] = $3;
    comment_db[_the_day] = sprintf( "%s=%imin", tl_should, $3 );
}

/^[0-9]+[[:blank:]]+(KOMMENTAR|COMMENT)[[:blank:]]+.*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    comment_db[_the_day] = $3;
}

/^(SUMMENBASIS|SUMBASE)[[:blank:]]+.[01-9]+[[:blank:]]+[01-9]+*$/ {
    consumed_line = 1;

    diffsumadder = 60 * $2 + $3
}

/^[0-9]+[[:blank:]]+(URLAUB|HOLIDAY)[[:blank:]]*$/ {
    consumed_line = 1;

    init_month( $1 );
    _the_day = "20"$1;
    # time_db[_the_day] = should_db[_the_day];
    holiday_db[_the_day] = 100;
    comment_db[_the_day] = tl_holiday;
}

/^[0-9]+[[:blank:]]+(URLAUBSTAGE|ADDHOLIDAY)[[:blank:]]+[0-9]+[[:blank:]]*$/ {
    consumed_line = 1;

    if (dbg) printf "ADDHOLIDAY: %i %i\n", $1, $3;
    holidays_begin[ "20"$1 ] += $3;
}

/^[[:blank:]]+[@].+$/ {
    consumed_line = 1;

    job_db_counter[active_day]++;
    _context = substr( $1, 2 );
    gsub( "^[[:blank:]]+[@][^[:blank:]]+[[:blank:]]*", "" );
    job_db[active_day,job_db_counter[active_day]] = sprintf( "%s@%s", _context, $0 );
}

/^[[:blank:]]+[^@[:blank:]].+$/ {
    consumed_line = 1;

    gsub( "^[[:blank:]]+", "" );
    job_db[active_day,job_db_counter[active_day]] = sprintf( "%s; %s",
	job_db[active_day,job_db_counter[active_day]], $0 );
}

/^[0-9][0-9]((1[3-9])|([2-9][0-9]))[[:blank:]]+/ {
  azat_error( sprintf( "%s: %s", "incorrect month found", $0 ));
}

/^[0-9][0-9][0-9][0-9]((3[2-9])|([4-9][0-9])|(00))[[:blank:]]+/ {
  azat_error( sprintf( "%s: %s", "incorrect day found", $0 ));
}

/^[0-9][0-9](0[1-9]|1[12])(0[1-9]|[12][0-9]|3[01])[[:blank:]]+/ {
  _now = $1;
  if (_now < last_parsed_date)
      azat_warning( sprintf( "out of order date found (%s < %s): %s",
			     _now, last_parsed_date, $0));
  last_parsed_date = _now;
}

/^[[:blank:]]*$/ { /* empty lines */
    consumed_line = 1;
}

/^#.*$/ { /* comment */
    consumed_line = 1;
}

/./ {
  if (consumed_line == 0)
    azat_warning( sprintf( "%s: %s", "unconsumed line", $0 ));
}

function output_month_donetable( ym,
				 the_day, ml, i, context, desc ) {
    if (dbg) printf "called output_month_donetable( %s )\n", ym;
    ml = date_get_monthlength( ym );
    if (latex) {
	printf "\\subsection{%s %s}\n",
	    tl_month_donetable, date_get_pretty_month( ym );
	printf "\\begin{longtable}{|lr|l|l|}\n";
	printf "\\hline\n";
	printf "%s & & %s & %s\\\\\n", tl_day, tl_context, tl_desc;
	printf "\\hline\n";
	printf "\\hline\n";
	printf "\\endhead\n";
	printf "\\hline\n";
	printf "\\endfoot\n";
    } else {
	printf "%s %s\n", tl_month_donetable, date_get_pretty_month( ym );
	printf "%s\t%s\t\t%s\n", tl_day, tl_context, tl_desc;
    }
    for (day = 1; day <= ml; day++) {
	the_day = sprintf( "%i%02i", ym, day );
	if (day > 1 && date_get_day_of_week( the_day ) == 1)
	    printf "\\hline\n\\hline\n";
	if (comment_db[the_day] != "") {
	    printf "%s & %s & %s & \\parbox[t]{10cm}{%s \\vspace{0.1cm}}\\\\\n",
		date_get_day_name_short( sprintf( "%i%02i", ym, day )),
		sprintf( "%i.", day ), tl_comment, comment_db[the_day];
	}
	for (i = 1; i <= job_db_counter[the_day]; i++) {
	    context = gensub( "[@].*$", "", "g", job_db[the_day,i] );
	    desc = gensub( "^[^@]+[@]", "", "1", job_db[the_day,i] );
	    gsub( "[&]", "\\\\&", desc );
            gsub( "[_]", "\\_", desc );
            gsub( "[%]", "\\%", desc );
	    gsub( "[\"]", "''", desc );
	    if (latex) {
		printf "%s & %s & %s & \\parbox[t]{10cm}{%s \\vspace{0.1cm}}\\\\\n",
		    date_get_day_name_short( sprintf( "%i%02i", ym, day )),
		    sprintf( "%i.", day ), context, desc;
	    } else
		printf "%i%02i\t%s\t\t%s\n", ym, day, context, desc;
	}
    }
    if (latex) {
	printf "\\end{longtable}\n\n";
	printf "\\newpage\n";
    } else {
	printf "\n";
    }
}

function output_month_timetable( ym,
				 the_day, ml, diffsum ) {
    if (dbg) printf "called output_month_timetable( %s )\n", ym;
    ml = date_get_monthlength( ym );
    if (latex) {
	printf "\\subsection{%s %s}\n\n",
	    tl_month_timetable, date_get_pretty_month( ym );
	printf "\\begin{tabular}{|lr|r|r|r|r|l|}\n";
	printf "\\hline\n";
	printf "%s & & %s & %s & %s & %s & %s\\\\\n",
	    tl_day, tl_work, tl_should, tl_diff_short, tl_sum, tl_comment;
	printf "\\hline\n";
	printf "\\hline\n";
    } else {
	printf "%s %s\n", tl_month_timetable, date_get_pretty_month( ym );
	printf "%s\t\t%s\t%s\t\t%s\t\t%s\t%s\n",
	    tl_day, tl_work, tl_should, tl_diff_short, tl_sum, tl_comment;
    }
    diffsum = 0;
    for (day = 1; day <= ml; day++) {
	the_day = sprintf( "%i%02i", ym, day );
	if (day > 1 && date_get_day_of_week( the_day ) == 1)
	    printf "\\hline\n\\hline\n";
	diffsum += time_db[the_day] - should_db[the_day];
	if (latex) {
	    printf "%s & %s & %s & %s & %s & %s & %s\\\\\n",
		date_get_day_name_short( sprintf( "%i%02i", ym, day )),
		sprintf( "%i.", day ),
		min2str( time_db[the_day] ),
		min2str( should_db[the_day] ),
		min2str( time_db[the_day] - should_db[the_day] ),
		min2str( diffsum ), comment_db[the_day];
	} else
	    printf "%i%02i\t%s\t%s\t%s\t%s\n", ym, day,
		min2str( time_db[the_day] ),
		min2str( should_db[the_day] ),
		min2str( time_db[the_day] - should_db[the_day] ),
		min2str( diffsum );
    }
    if (latex) {
	printf "\\hline\n";
	printf "\\end{tabular}\n\n";
	printf "\\newpage\n\n";
    } else {
	printf "\n";
    }
}

function output_month_summary( ym ) {
    if (dbg) printf "called output_month_summary( %s )\n", ym;
    if (latex) {
	printf "%s & %s & %s & %s & %s & %s & %i & %i\\\\\n",
	    date_get_year( ym ),
	    date_get_month_name_long( ym ), min2str( month_time[ym] ),
	    min2str( month_should[ym] ), min2str( month_diff[ym] ),
	    min2str( month_diffsum[ym] ), holidays_begin[ym], holidays_end[ym];
    } else {
	printf "%s\t%s\t%s\t%s\t%s\t%i\t\t\t\t\t%i\n",
	    date_get_year( ym ),
	    date_get_month_name_long( ym ), min2str( month_time[ym] ),
	    min2str( month_should[ym] ), min2str( month_diff[ym] ),
	    min2str( month_diffsum[ym] ), holidays_begin[ym], holidays_end[ym];
    }
}

function output_month_summaries( i ) {
    if (latex) {
	printf "\\section{%s}\n\n", tl_month_overview;
	if (useddiffsumadder > 0) {
	  printf "Manueller Stundenübertrag: %s\\\\\n",
	    min2str( useddiffsumadder)
	}
	printf "\\begin{tabular}{|l|l|r|r|r|r|r|r|}\n";
	printf "\\hline\n";
	printf "& & & & & & %s & %s\\\\\n",
	    tl_holidays, tl_holidays;
	printf "%s & %s & %s & %s & %s & %s & %s & %s\\\\\n",
	    tl_year, tl_month, tl_is, tl_should, tl_diff, tl_sum, tl_before,
	    tl_after;
	printf "\\hline\n";
    } else {
	printf "%s:\n", tl_month_overview;
	printf "%s\t%s\t%s\t\t%s\t%s\t%s\t\t%s %s\t%s %s\n",
	    tl_year, tl_month, tl_is, tl_should, tl_diff, tl_sum,
	    tl_holidays, tl_before, tl_holidays, tl_after;
    }
    for (i = 1; i <= month_counter; i++)
	output_month_summary( month_db[i] );
    if (latex) {
	printf "\\hline\n";
	printf "\\end{tabular}\n"
	printf "\\newpage\n\n";
    } else {
	printf "\n";
    }
}

function output_header() {
    if (latex) {
      printf "\\markright{%s %s}\n", tl_title, worker_name;
	printf "\\title{%s %s}\n", tl_title, worker_name;
	printf "\\author{%s \\texttt{<}%s\\texttt{>}}\n",
	    worker_name, worker_mail;
	printf "\\maketitle\n";
	if (error_counter > 0) {
	  printf "\\vspace{2cm}\n";
	  printf "{\\huge AZAT Warnings/Errors found:}\n\n";
	  printf "\\vspace{1cm}\n";
	  for (i = 0; i < error_counter; i++)
	    printf "%s\n\n", errors[i];
	}
	if (!suppress_ad) {
	    printf "\\vfill\n";
	    printf "\\begin{center}\n";
	    printf "\\textsl{\\textsf{created by azat ";
	    printf azat_version;
	    printf " - https://git.inf2.tuwien.ac.at/inf2/sw/azat}}\n";
	    printf "\\end{center}\n";
	}
	printf "\\newpage\n";
	printf "\\tableofcontents\n\\newpage\n\n";
    } else {
	printf "%s %s\n\n", tl_title, worker_name;
    }
}

BEGIN {
    lang_init();
}

END {
    asort( month_db );
    if (dbg) {
	printf "dumping monthdb begin\n";
	for (x in month_db) printf " month_db[%s]=%s\n", x, month_db[x];
	printf "dumping monthdb end\n";
    }
    for (i = 1; i <= month_counter; i++) {
	calc_month_holidays( month_db[i], month_db[i-1] );
	calc_month_timetable( month_db[i], month_db[i-1], diffsumadder );
	useddiffsumadder = diffsumadder;
        diffsumadder = 0;
    }

    output_header();
    output_month_summaries();

    if (latex)
	printf "\\section{%s}\n\n", tl_month_timetable_title;
    for (i = 1; i <= month_counter; i++)
	output_month_timetable( month_db[i] );

    if (add_work_overview) {
	if (latex)
	    printf "\\section{%s}\n\n", tl_month_donetable_title;
	for (i = 1; i <= month_counter; i++)
	    output_month_donetable( month_db[i] );
    }
}
